import logging
from functools import lru_cache

import gnupg
import pkg_resources
# noinspection PyPackageRequirements
from django.core.signing import Signer, BadSignature
# noinspection PyPackageRequirements
from django.utils.crypto import constant_time_compare

__author__ = 'flanker'
logger = logging.getLogger('django.requests')
GPG_CONF_FILENAME = pkg_resources.resource_filename('moneta', 'templates/gpg.conf')


@lru_cache()
def get_gpg():
    # noinspection PyPackageRequirements
    from django.conf import settings
    gpg = gnupg.GPG(homedir=settings.GNUPG_HOME, binary=settings.GNUPG_PATH,
                    secring='secring.gpg', keyring='pubring.gpg',
                    options=['--options', GPG_CONF_FILENAME])
    return gpg


class GPGSigner(Signer):
    def __init__(self, key=None, sep=':', salt=None):
        # noinspection PyPackageRequirements
        from django.conf import settings
        super().__init__(sep=sep, salt=salt)
        self.key = str(key or settings.GNUPG_KEYID)

    def signature(self, value):
        # noinspection PyPackageRequirements
        from django.conf import settings
        # noinspection PyUnresolvedReferences
        return str(get_gpg().sign(value, default_key=self.key, detach=True, digest_algo=settings.GNUPG_DIGEST_ALGO))

    def sign_file(self, fd, detach=True):
        # noinspection PyPackageRequirements
        from django.conf import settings
        # noinspection PyUnresolvedReferences
        return str(get_gpg().sign(fd, default_key=self.key, detach=detach, clearsign=not detach,
                                  digest_algo=settings.GNUPG_DIGEST_ALGO))

    def export_key(self):
        # noinspection PyUnresolvedReferences
        return get_gpg().export_keys(self.key)

    def sign(self, value):
        value = str(value)
        return str('%s%s%s') % (value, self.sep, self.signature(value))

    def unsign(self, signed_value):
        signed_value = str(signed_value)
        if self.sep not in signed_value:
            raise BadSignature('No "%s" found in value' % self.sep)
        value, sig = signed_value.rsplit(self.sep, 1)
        if constant_time_compare(sig, self.signature(value)):
            return str(value)
        raise BadSignature('Signature "%s" does not match' % sig)

    # noinspection PyMethodMayBeStatic
    def verify(self, value):
        # noinspection PyUnresolvedReferences
        return get_gpg().verify(value)

    # noinspection PyMethodMayBeStatic
    def verify_file(self, fd, path_to_data_file=None):
        # noinspection PyUnresolvedReferences
        return get_gpg().verify_file(fd, path_to_data_file)



# Generated by Django 1.10.6 on 2017-03-06 21:02

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('repository', '0002_auto_20161226_2320'),
    ]

    operations = [
        migrations.AlterField(
            model_name='elementsignature',
            name='method',
            field=models.CharField(choices=[('gpg', 'GnuPG'), ('x509', 'x509 (openSSL/libreSSL)')], db_index=True, max_length=10, verbose_name='méthode de signature'),
        ),
    ]



import base64
import gzip
import io
import json
import subprocess
import tarfile
import tempfile
import zlib
from distutils.version import LooseVersion

# noinspection PyPackageRequirements
import yaml
from django.conf import settings
from django.conf.urls import url
from django.core.files.uploadedfile import UploadedFile
from django.urls import reverse
from django.http import HttpResponse, Http404
from django.shortcuts import get_object_or_404
from django.template.response import TemplateResponse
from django.utils.translation import ugettext as _
from rubymarshal.classes import UsrMarshal
from rubymarshal.writer import write
# noinspection PyPackageRequirements
from yaml import MappingNode
# noinspection PyPackageRequirements
from yaml.composer import Composer
# noinspection PyPackageRequirements
from yaml.constructor import Constructor
# noinspection PyPackageRequirements
from yaml.parser import Parser
# noinspection PyPackageRequirements
from yaml.reader import Reader
# noinspection PyPackageRequirements
from yaml.resolver import Resolver
# noinspection PyPackageRequirements
from yaml.scanner import Scanner

from moneta.repositories.aptitude import Aptitude
from moneta.repositories.base import RepositoryModel
from moneta.repository.models import ArchiveState, Element, Repository, storage
from moneta.templatetags.moneta import moneta_url
from moneta.views import sendpath

__author__ = 'flanker'


class RubyMarshal(UsrMarshal):
    def __init__(self, values):
        super().__init__('Gem::' + self.__class__.__name__, values=values)


class Version(RubyMarshal):
    @property
    def version(self):
        return self.values['version']


class Specification(RubyMarshal):
    @property
    def name(self):
        return self.values['name']

    @property
    def version(self) -> Version:
        return self.values['version']

    @property
    def platform(self):
        return self.values['platform']

    @property
    def description(self):
        return self.values['description']

    @property
    def rubygems_version(self):
        return self.values['rubygems_version']

    @property
    def summary(self):
        return self.values['summary']


class Dependency(RubyMarshal):
    @property
    def name(self):
        return self.values['name']

    @property
    def requirement(self):
        return self.values['requirement']

    @property
    def type(self):
        return self.values['type']

    @property
    def prerelease(self):
        return self.values['prerelease']

    @property
    def version_requirements(self):
        return self.values['version_requirements']


class Requirement(RubyMarshal):
    @property
    def requirements(self):
        return self.values['requirements']


class RubyConstructor(Constructor):
    def construct_version(self, node, deep=False):
        if isinstance(node, MappingNode):
            self.flatten_mapping(node)
        return Version(super(RubyConstructor, self).construct_mapping(node, deep=deep))

    def construct_specification(self, node, deep=False):
        if isinstance(node, MappingNode):
            self.flatten_mapping(node)
        return Specification(super(RubyConstructor, self).construct_mapping(node, deep=deep))

    def construct_dependency(self, node, deep=False):
        if isinstance(node, MappingNode):
            self.flatten_mapping(node)
        return Dependency(super(RubyConstructor, self).construct_mapping(node, deep=deep))

    def construct_requirement(self, node, deep=False):
        if isinstance(node, MappingNode):
            self.flatten_mapping(node)
        return Requirement(super(RubyConstructor, self).construct_mapping(node, deep=deep))


RubyConstructor.add_constructor('!ruby/object:Gem::Specification', RubyConstructor.construct_specification)
RubyConstructor.add_constructor('!ruby/object:Gem::Version', RubyConstructor.construct_version)
RubyConstructor.add_constructor('!ruby/object:Gem::Dependency', RubyConstructor.construct_dependency)
RubyConstructor.add_constructor('!ruby/object:Gem::Requirement', RubyConstructor.construct_requirement)


class RubyLoader(Reader, Scanner, Parser, Composer, RubyConstructor, Resolver):
    def __init__(self, stream):
        Reader.__init__(self, stream)
        Scanner.__init__(self)
        Parser.__init__(self)
        Composer.__init__(self)
        RubyConstructor.__init__(self)
        Resolver.__init__(self)


class RubyGem(Aptitude):
    verbose_name = _('Gem repository for Ruby packages')
    storage_uid = '71610000-0000-0000-0000-%012d'
    archive_type = 'rubygem'

    def finish_element(self, element: Element, states: list):
        """
        Called after the .save() operations, with all states associated to this new element.
        :param element: Element
        :param states: list of ArchiveState
        """
        RepositoryModel.finish_element(self, element, states)
        # self.generate_indexes(element.repository)

    def is_file_valid(self, uploaded_file: UploadedFile):
        if not uploaded_file.name.endswith('.gem'):
            return False
        try:
            fileobj = uploaded_file.file
            fd = tarfile.open(fileobj=fileobj, mode='r|')
            names = fd.getnames()
            fd.close()
        except tarfile.TarError:
            return False
        if 'data.tar.gz' not in names or 'metadata.gz' not in names:
            return False
        return True

    def update_element(self, element):
        """
        Extract some informations from element to prepare the repository
        :param element: Element to add to the repository
        :return: Unicode string containing meta-data

        ar -x control.tar.gz
        tar -xf control.tar.gz control
        """
        archive_file = storage(settings.STORAGE_ARCHIVE).get_file(element.archive_key, '')
        gem_fd = tarfile.open(fileobj=archive_file, mode='r')
        metadata_fd = gem_fd.extractfile('metadata.gz')
        metadata_gz_content = metadata_fd.read()
        metadata_bytes = gzip.decompress(metadata_gz_content)
        gem_fd.close()
        data = yaml.load(io.BytesIO(metadata_bytes), Loader=RubyLoader)
        for key, attr in (('name', 'archive'), ('homepage', 'official_link'), ('summary', 'long_description'),
                          ('name', 'name')):
            if key in data.values:
                setattr(element, attr, data.values[key])
        element.version = data.values['version'].version
        p = subprocess.Popen(['ruby', '-e', 'puts Marshal.dump (Gem::Specification.from_yaml(ARGF.read))'],
                             stdin=subprocess.PIPE, stdout=subprocess.PIPE)
        stdout, stderr = p.communicate(metadata_bytes)
        extra_data = {'yaml': metadata_bytes.decode('utf-8'), 'marshal': base64.b64encode(stdout).decode('utf-8')}
        element.extra_data = json.dumps(extra_data)

    @staticmethod
    def marshal_data(elt: Element) -> bytes:
        marshal = json.loads(elt.extra_data)['marshal']
        return base64.b64decode(marshal.encode('utf-8'))

    @staticmethod
    def python_data(elt: Element) -> Specification:
        data = json.loads(elt.extra_data)['yaml']
        return yaml.load(io.BytesIO(data.encode('utf-8')), Loader=RubyLoader)

    @staticmethod
    def yaml_data(elt: Element) -> Specification:
        return json.loads(elt.extra_data)['yaml']

    def public_url_list(self):
        """
        Return a list of URL patterns specific to this repository
        Sample recognized urls:
            *

        :return: a patterns as expected by django

        """
        src_pattern_list = [(r'(?P<filename>(specs\.4\.8|prerelease_specs\.4\.8|latest_specs\.4\.8|Marshal\.4\.8|'
                             r'versions\.list|names\.list)(\.gz)?)', 'specs', 'specs'),
                            (r'gems/(?P<filename>.+)', 'download', 'download'),
                            (r'specs/(?P<filename>.+)\.gemspec', 'gem_specs', 'gem_specs'),
                            (r'quick/Marshal\.4\.8/(?P<filename>.+)\.gemspec(?P<compression>(\.rz|))',
                             'quick_gem_specs', 'quick_gem_specs'),
                            (r'', 'index', 'index'),
                            ]
        pattern_list = []
        for pattern, view, name in src_pattern_list:
            pattern_list.append(
                url(r"^(?P<rid>\d+)/(?P<repo_slug>[\w\-\._]*)/a/%s$" % pattern, self.wrap_view(view), name=name)
            )
            pattern_list.append(
                url(r"^(?P<rid>\d+)/(?P<repo_slug>[\w\-\._]*)/s/(?P<state_slug>[\w\-\._]+)/%s$" % pattern,
                    self.wrap_view(view), name=name)
            )
        pattern_list += [
            url(r"^(?P<rid>\d+)/$", self.wrap_view('index'), name="index"),
        ]
        return pattern_list

    def specs(self, request, rid, repo_slug, state_slug=None, filename='specs.4.8.gz'):
        # noinspection PyUnusedLocal
        repo_slug = repo_slug
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        if state_slug:
            filename = 'specs/%(slug)s/%(filename)s' % {'slug': state_slug, 'filename': filename, }
        else:
            filename = 'specs/%(filename)s' % {'filename': filename, }
        uid = self.storage_uid % repo.pk
        key = storage(settings.STORAGE_CACHE).uid_to_key(uid)
        return sendpath(settings.STORAGE_CACHE, key, filename, 'application/gzip')

    def quick_gem_specs(self, request, rid, repo_slug, state_slug=None, filename=None, compression=''):
        name, sep, version = filename.rpartition('-')
        if sep != '-':
            raise Http404
        # noinspection PyUnusedLocal
        repo_slug = repo_slug
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        base_query = Element.objects.filter(repository=repo)
        if state_slug:
            state = get_object_or_404(ArchiveState, repository=repo, name=state_slug)
            base_query = base_query.filter(states=state)
        element = get_object_or_404(base_query, name=name, version=version)
        content = self.marshal_data(element)
        if compression == '.rz':
            content = zlib.compress(content)
        return HttpResponse(content, content_type='application/gzip')

    def gem_specs(self, request, rid, repo_slug, state_slug=None, filename=None):
        name, sep, version = filename.rpartition('-')
        if sep != '-':
            raise Http404
        # noinspection PyUnusedLocal
        repo_slug = repo_slug
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        base_query = Element.objects.filter(repository=repo)
        if state_slug:
            state = get_object_or_404(ArchiveState, repository=repo, name=state_slug)
            base_query = base_query.filter(states=state)
        element = get_object_or_404(base_query, name=name, version=version)
        return HttpResponse(self.yaml_data(element), content_type='text/yaml')

    def download(self, request, rid, repo_slug, state_slug=None, filename=None):
        # noinspection PyUnusedLocal
        repo_slug = repo_slug
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        base_query = Element.objects.filter(repository=repo)
        if state_slug:
            state = get_object_or_404(ArchiveState, repository=repo, name=state_slug)
            base_query = base_query.filter(states=state)
        element = get_object_or_404(base_query, filename=filename)
        from moneta.views import get_file
        return get_file(request, element.pk)

    def index(self, request, rid, repo_slug=None, state_slug=None):
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        states = list(ArchiveState.objects.filter(repository=repo).order_by('name'))
        template_values = {'repo': repo, 'states': states, 'upload_allowed': repo.upload_allowed(request),
                           'admin_allowed': repo.admin_allowed(request), }
        view_name = moneta_url(repo, 'index')
        tab_infos = [
            (reverse(view_name, kwargs={'rid': repo.id, 'repo_slug': repo.slug}), states,
             ArchiveState(name=_('All states'), slug='all-states')),
        ]
        for state in states:
            tab_infos.append(
                (reverse(view_name, kwargs={'rid': repo.id, 'repo_slug': repo.slug, 'state_slug': state.slug}),
                 [state], state)
            )
        template_values['tab_infos'] = tab_infos
        return TemplateResponse(request, 'repositories/ruby/index.html', template_values)

    def generate_indexes(self, repo: Repository, states=None, validity=365):
        if states is None:
            states = list(ArchiveState.objects.filter(repository=repo))
        base_query = Element.objects.filter(repository=repo).prefetch_related('states')
        state_infos = [(None, None)] + [(state.slug, state.pk) for state in states]
        all_elements = {state_info[0]: [] for state_info in state_infos}
        all_specs = {state_info[0]: [] for state_info in state_infos}
        last_elements = {state_info[0]: {} for state_info in state_infos}
        for element in base_query:
            element_spec = self.python_data(element)
            assert isinstance(element_spec, Specification)
            element_data = [element.archive, element_spec.version, element_spec.platform]
            all_specs[None].append(element_spec)
            if 'date' in element_spec.values:
                del element_spec.values['date']
            all_elements[None].append(element_data)
            last_elements[None].setdefault(element_spec.name, []).append((element_spec.version.version,
                                                                          element_spec.platform))
            for state in element.states.all():
                all_elements[state.slug].append(element_data)
                all_specs[state.slug].append(element_spec)
                last_elements[state.slug].setdefault(element_spec.name, []).append((element_spec.version.version,
                                                                                    element_spec.platform))

        for state_info in state_infos:
            folder_name = 'specs/' if state_info[0] is None else 'specs/%(slug)s' % {'slug': state_info[0]}

            last_elements_by_state = []
            for name, versions in last_elements[state_info[0]].items():
                try:
                    versions.sort(key=lambda x: LooseVersion(x[0]))
                    last_elements_by_state.append([name, Version([versions[-1][0]]), versions[-1][1].encode('utf-8')])
                except ValueError:
                    continue
            names = list(last_elements[state_info[0]].keys())
            names.sort()

            def versions_write_fn(file_obj):
                for y in names:
                    version_txt = ('%s %s\n' % (y, ','.join([z[0] for z in last_elements[state_info[0]][y]])))
                    file_obj.write(version_txt.encode('utf-8'))

            def names_write_fn(file_obj):
                for y in names:
                    file_obj.write(('%s\n' % y).encode('utf-8'))

            self.__write_file(repo, '%s/specs.4.8' % folder_name, lambda x: write(x, all_elements))
            self.__write_file(repo, '%s/latest_specs.4.8' % folder_name, lambda x: write(x, last_elements_by_state))
            self.__write_file(repo, '%s/prerelease_specs.4.8' % folder_name, lambda x: write(x, []))
            self.__write_file(repo, '%s/Marshal.4.8' % folder_name, lambda x: write(x, all_specs))
            self.__write_file(repo, '%s/versions.list' % folder_name, versions_write_fn)
            self.__write_file(repo, '%s/names.list' % folder_name, names_write_fn)

    def __write_file(self, repo, dest_filename, write_function):
        uid = self.storage_uid % repo.pk
        key = storage(settings.STORAGE_CACHE).uid_to_key(uid)
        plain_file = tempfile.NamedTemporaryFile(mode='w+b', dir=settings.FILE_UPLOAD_TEMP_DIR, delete=False)
        gz_plain_file = tempfile.NamedTemporaryFile(mode='w+b', dir=settings.FILE_UPLOAD_TEMP_DIR, delete=False)
        gz_file = gzip.open(gz_plain_file, 'wb')
        write_function(plain_file)
        plain_file.flush()
        plain_file.seek(0)
        for block in iter(lambda: plain_file.read(8192), b''):
            gz_file.write(block)
        gz_file.close()
        gz_plain_file.close()
        storage(settings.STORAGE_CACHE).import_filename(plain_file.name, key, dest_filename)
        storage(settings.STORAGE_CACHE).import_filename(gz_plain_file.name, key, dest_filename + '.gz')



import os
import pwd
import re
import stat
from argparse import ArgumentParser

import datetime
from django.conf import settings
from django.core.management.base import BaseCommand

from moneta.repository.signing import get_gpg

__author__ = 'flanker'

if settings.ADMINS and len(settings.ADMINS[0]) == 2:
    default_email = settings.ADMINS[0][1]
else:
    default_email = 'moneta@19pouces.net'
GPG = get_gpg()


class Command(BaseCommand):
    args = '<generate|show|export>'
    help = """command=generate: Create a new GPG key
    command=show: Show existing GPG keys
    command=export: export GPG key"""

    def add_arguments(self, parser):
        assert isinstance(parser, ArgumentParser)
        parser.add_argument('gpg_command', action='store', default=None, choices=['generate', 'show', 'export']),
        parser.add_argument('--type', action='store', default='RSA', help='Key type, RSA (default) or DSA.',
                            choices=('RSA', 'DSA')),
        parser.add_argument('--length', action='store', default=2048, help='Key length (default 2048).',
                            type=int),
        parser.add_argument('--name', action='store', default='Moneta GNUPG key', help='Name of the key'),
        parser.add_argument('--email', action='store', default=default_email, help='Email address for the user.'),
        parser.add_argument('--years', action='store', default='10y',
                            help='Expiration date, in number of years (like "10y") or days (like "10d").'),
        parser.add_argument('--absent', action='store_true',
                            default=False, help='Generate keys only when no key already exists.'),
        parser.add_argument('--onlyid', action='store_true', default=False,
                            help='Only display the ID of the keys (that is the ID expected in the config gile).'),

    def handle(self, *args, **options):
        command = options['gpg_command']
        if command == 'generate':
            if options['absent'] and len(GPG.list_keys(False)) > 0:
                return
            year_matcher = re.match(r'^(\d+)y$', options['years'])
            day_matcher = re.match(r'^(\d+)d$', options['years'])
            today = datetime.date.today()
            if not year_matcher and not day_matcher:
                year_matcher = re.match(r'^(\d+)y$', '10y')
            if year_matcher:
                expire = '%s-%02d-%02d' % (today.year + int(year_matcher.group(1)), today.month, today.day)
                if today.month == 2 and today.day >= 28:
                    expire = '%s-%02d-%02d' % (today.year + int(year_matcher.group(1)), today.month, 28)
            else:
                today += datetime.timedelta(days=int(day_matcher.group(1)))
                expire = '%s-%02d-%02d' % (today.year, today.month, today.day)
            input_data = 'Key-Type: %s\n' % options['type']
            input_data += "Key-Length: %d\n" % options['length']
            input_data += "Name-Real: %s\n" % options['name']
            input_data += "Expire-Date: %s\n" % expire
            input_data += "Name-Email: %s\n" % options['email']
            input_data += '%no-protection\n'
            input_data += '%commit\n'
            key = GPG.gen_key(input_data)
            self.stdout.write("Fingerprint %s" % key)
        elif command == 'show':
            if options['onlyid']:
                for key in GPG.list_keys(False):
                    self.stdout.write("{keyid}".format(**key))
            else:
                self.stdout.write("Available keys:")
                for key in GPG.list_keys(False):
                    self.stdout.write("id (GNUPG_KEYID) : {keyid}, "
                                      "length : {length}, fingerprint : {fingerprint}".format(**key))
        elif command == 'export':
            self.stdout.write(GPG.export_keys(settings.GNUPG_KEYID))
        self.check_rights()

    def check_rights(self):
        os_stat = os.stat(settings.GNUPG_HOME)
        must_apply_rights = self.check_mode(os_stat, stat.S_IRUSR | stat.S_IXUSR | stat.S_IWUSR)
        must_apply_owners = self.check_owner(os_stat)
        for root, dirnames, filenames in os.walk(settings.GNUPG_HOME):
            for filename in filenames:
                if 'gpg-agent' in filename or 'pub' in filename or 'daemon' in filename:
                    continue
                os_stat = os.stat(os.path.join(root, filename))
                must_apply_rights |= self.check_mode(os_stat, stat.S_IRUSR | stat.S_IWUSR)
                must_apply_owners |= self.check_owner(os_stat)
            for dirname in dirnames:
                os_stat = os.stat(os.path.join(root, dirname))
                must_apply_rights |= self.check_mode(os_stat, stat.S_IRUSR | stat.S_IXUSR | stat.S_IWUSR)
                must_apply_owners |= self.check_owner(os_stat)

        if must_apply_rights:
            self.stderr.write('Invalid permissions. You should run the following commands to fix them')
            self.stderr.write('chmod 700 "%s"' % settings.GNUPG_HOME)
            self.stderr.write('find "%s" -type d -exec chmod 700 {} \\;' % settings.GNUPG_HOME)
            self.stderr.write('find "%s" -type f -exec chmod 600 {} \\;' % settings.GNUPG_HOME)
        if must_apply_owners:
            user = pwd.getpwuid(os.getuid())
            if user:
                self.stderr.write('Invalid file owners. You should run the following command to fix them')
                self.stderr.write('chown -R "%s" %s' % (settings.GNUPG_HOME, user[0]))

    @staticmethod
    def check_owner(os_stat):
        return os_stat.st_uid != os.getuid()

    @staticmethod
    def check_mode(os_stat, expected):
        mask = stat.S_IRWXU | stat.S_IRWXG | stat.S_IRWXO
        current_mode = os_stat.st_mode & mask
        return bool((current_mode & mask) - (current_mode & expected)) or (current_mode & expected != expected)



# https://www.nopsec.com/news-and-resources/blog/2015/3/27/private-vagrant-box-hosting-easy-versioning/
import json
import tarfile

from django.conf import settings
from django.conf.urls import url
from django.core.files.uploadedfile import UploadedFile
from django.urls import reverse
from django.http import HttpRequest
from django.http.response import JsonResponse
from django.shortcuts import get_object_or_404
from django.template.response import TemplateResponse
from django.utils.translation import ugettext as _

from moneta.repositories.base import RepositoryModel
from moneta.repository.models import ArchiveState, Element, storage
from moneta.repository.models import Repository
from moneta.templatetags.moneta import moneta_url
from moneta.views import get_file

__author__ = 'Matthieu Gallet'


class Vagrant(RepositoryModel):
    verbose_name = _('Vagrant custom box catalog')
    storage_uid = 'vagrant-0000-0000-0000-%012d'
    archive_type = 'vagrant'
    index_html = 'repositories/vagrant/index.html'

    def is_file_valid(self, uploaded_file: UploadedFile):
        name = uploaded_file.name
        if not name.endswith('.box'):
            return False
        # noinspection PyBroadException
        try:
            tarfile.open(fileobj=uploaded_file.file, mode='r', name=None)
        except Exception:
            return False
        return True

    def update_element(self, element):
        """
        Extract some informations from element to prepare the repository
        :param element: Element to add to the repository
        :return: Unicode string containing meta-data
        """
        if element.archive:
            element.name = element.archive.rpartition('.')[2]
        archive_file = storage(settings.STORAGE_ARCHIVE).get_file(element.archive_key)
        compressed_file = tarfile.open(name=None, fileobj=archive_file, mode='r')
        all_names = {x for x in compressed_file.getnames()}
        provider = 'virtualbox'
        if 'metadata.json' in all_names:
            metadata_file = compressed_file.extractfile('metadata.json')
            metadata = json.loads(metadata_file.read().decode('utf-8'))
            provider = metadata['provider']
        element.extra_data = json.dumps({'provider': provider})

    def finish_element(self, element: Element, states: list):
        """
        Called after the .save() operations, with all states associated to this new element.
        Remove previous versions from target states
        :param element: Element
        :param states: list of ArchiveState
        """
        # remove previous versions from the given states:
        # noinspection PyUnresolvedReferences
        Element.states.through.objects.exclude(element__version=element.version) \
            .filter(archivestate__in=states, element__archive=element.archive).delete()
        super().finish_element(element, states)

    def public_url_list(self):
        # vagrant.org.com/org/base-trusty64.json
        # vagrant.org.com/org/base-trusty64/version/1.0.0/virtualbox/base-trusty64-1.0.0-virtualbox.json

        return [
            # https://atlas.hashicorp.com/bento/boxes/centos-7.1/versions/2.2.2/providers/vmware_desktop.box
            url(r"^(?P<rid>\d+)/(?P<repo_slug>[\w\-\._]+)/boxes/(?P<eid>\d+)-(?P<archive>[\w\-\._]+)/"
                r"versions/(?P<version>[\w\-\._]+)/providers/(?P<provider>[\w\-\._]+)\.box",
                self.wrap_view('get_box'), name='get_box'),
            url(r"^(?P<rid>\d+)/(?P<repo_slug>[\w\-\._]*)/s/(?P<state_slug>[\w\-\._]+)/(?P<archive>[^/]+)\.json$",
                self.wrap_view('archive_json'), name='archive_json'),
            url(r"^(?P<rid>\d+)/s(?P<state_slug>[\w\-\._]*)\.html$", self.wrap_view('index'), name="index"),
            url(r"^(?P<rid>\d+)\.html$", self.wrap_view('index'), name="index"),
        ]

    def get_box(self, request: HttpRequest, rid, repo_slug, eid, archive, version, provider):
        # noinspection PyUnusedLocal
        repo_slug = repo_slug
        # noinspection PyUnusedLocal
        archive = archive
        # noinspection PyUnusedLocal
        provider = provider
        # noinspection PyUnusedLocal
        version = version
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        elements = list(Element.objects.filter(repository=repo, pk=eid)[0:1])
        if not elements:
            return JsonResponse({'errors': ['Not found', ], })
        return get_file(request, eid, element=elements[0])

    def archive_json(self, request: HttpRequest, rid, repo_slug, state_slug=None, archive=None):
        # noinspection PyUnusedLocal
        repo_slug = repo_slug
        result = self.get_providers_by_version(request, rid, state_slug, archive)
        return JsonResponse(result)

    # noinspection PyUnusedLocal
    def index(self, request, rid, repo_slug=None, state_slug=''):
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        base_query = Element.objects.filter(repository=repo)
        if state_slug:
            state = get_object_or_404(ArchiveState, repository=repo, name=state_slug)
            base_query = base_query.filter(states=state)
        base_query = base_query.order_by('archive', 'version')
        element_infos = []
        for element in base_query:
            if not element.extra_data:
                continue
            provider = json.loads(element.extra_data)['provider']
            element_infos.append(
                (element.name,
                 reverse('repository:vagrant:get_box',
                         kwargs={'rid': rid, 'repo_slug': repo.slug, 'eid': element.id, 'provider': provider,
                                 'archive': element.archive, 'version': element.version}),
                 element.sha1)
            )
        states = [state for state in ArchiveState.objects.filter(repository=repo).order_by('name')]
        tab_infos = []
        # list of (relative URL, name, list of states, state_slug)
        tab_infos += [(reverse('repository:vagrant:index',
                               kwargs={'rid': repo.id, 'state_slug': state.slug}),
                       state, [state], state.slug)
                      for state in states]
        tab_infos += [(reverse('repository:vagrant:index', kwargs={'rid': repo.id, 'state_slug': ''}),
                      ArchiveState(name=_('All boxes'), slug='all-states'), states, ''), ]

        template_values = {'repo': repo, 'states': states, 'upload_allowed': repo.upload_allowed(request),
                           'state_slug': state_slug, 'elements': element_infos,
                           'index_url': reverse(moneta_url(repo, 'index'), kwargs={'rid': repo.id, }),
                           'tab_infos': tab_infos, 'admin_allowed': repo.admin_allowed(request), }
        return TemplateResponse(request, self.index_html, template_values)

    def get_providers_by_version(self, request, rid, state_slug, archive):
        repo = get_object_or_404(Repository.reader_queryset(request), id=rid, archive_type=self.archive_type)
        base_query = Element.objects.filter(repository=repo, archive=archive)
        if state_slug:
            state = get_object_or_404(ArchiveState, repository=repo, name=state_slug)
            base_query = base_query.filter(states=state)
        versions = {}
        for element in base_query:
            metadata = json.loads(element.extra_data)
            versions.setdefault(element.version, []).append({
                'name': metadata['provider'],
                'url': '%s%s' % (settings.SERVER_BASE_URL[:-1], element.get_direct_link()),
                'checksum_type': 'sha1',

                'checksum': element.sha1})
        result = {'name': archive, 'description': _('This box contains %(name)s') % {'name': archive},
                  'versions': [{'version': k, 'status': 'active', 'providers': v} for (k, v) in versions.items()]}
        return result
# Vagrant.configure(2) do |config|
#
#   config.vm.box = "{$relativePathInfo|escape}"
#   config.vm.box_url = '<a href="{$CATALOG_URI|escape}{$pathInfo|escape}">{$CATALOG_URI|escape}{$pathInfo|escape}</a>'
#
#   # Whatever other config stuff you want to do
# end



# Generated by Django 1.10.4 on 2016-12-26 22:20

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('repository', '0001_initial'),
    ]

    operations = [
        migrations.AlterField(
            model_name='repository',
            name='on_index',
            field=models.BooleanField(db_index=True, default=True, verbose_name="Afficher sur l'index public ?"),
        ),
    ]



from djangofloor.conf.fields import bool_setting, CharConfigField
from djangofloor.conf.mapping import BASE_MAPPING, REDIS_MAPPING, AUTH_MAPPING, SENDFILE_MAPPING, ALLAUTH_MAPPING

__author__ = 'flanker'


def x_accel_converter(value):
    if bool_setting(value):
        return [('{MEDIA_ROOT}/', '/p/get/'),
                ('{MEDIA_ROOT}/', '/a/get/'), ]
    return []


INI_MAPPING = BASE_MAPPING + AUTH_MAPPING + ALLAUTH_MAPPING + REDIS_MAPPING + SENDFILE_MAPPING + [
    CharConfigField('gnupg.home', 'GNUPG_HOME', help_str='Path of the GnuPG secret data'),
    CharConfigField('gnupg.keyid', 'GNUPG_KEYID', help_str='ID of the GnuPG key'),
    CharConfigField('gnupg.path', 'GNUPG_PATH', help_str='Path of the gpg binary'),

]



from django.contrib import messages
from django.template.response import TemplateResponse
from django.utils.translation import ugettext_lazy as _


def public_check(request):
    messages.success(request, _('You can access to this page.'))
    return TemplateResponse(request, 'core/empty.html', {})


def private_check(request):
    if request.user.is_anonymous:
        messages.error(request, _('You are not authenticated.'))
    else:
        messages.success(request, _('You can access to this page and you are authenticated.'))
    return TemplateResponse(request, 'core/empty.html', {})



# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-11-04 12:00
from __future__ import unicode_literals

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion


class Migration(migrations.Migration):

    dependencies = [
        ('repository', '0004_auto_20171104_1257'),
    ]

    operations = [
        migrations.AlterField(
            model_name='archivestate',
            name='author',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL, verbose_name='Author'),
        ),
        migrations.AlterField(
            model_name='archivestate',
            name='creation',
            field=models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Creation date'),
        ),
        migrations.AlterField(
            model_name='archivestate',
            name='modification',
            field=models.DateTimeField(auto_now=True, db_index=True, verbose_name='Modification date'),
        ),
        migrations.AlterField(
            model_name='archivestate',
            name='name',
            field=models.CharField(db_index=True, max_length=100, verbose_name='Name'),
        ),
        migrations.AlterField(
            model_name='archivestate',
            name='repository',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='repository.Repository', verbose_name='Repository'),
        ),
        migrations.AlterField(
            model_name='archivestate',
            name='slug',
            field=models.SlugField(max_length=100, verbose_name='Slug'),
        ),
        migrations.AlterField(
            model_name='element',
            name='author',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL, verbose_name='Author'),
        ),
        migrations.AlterField(
            model_name='element',
            name='creation',
            field=models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Creation date'),
        ),
        migrations.AlterField(
            model_name='element',
            name='extension',
            field=models.CharField(blank=True, db_index=True, default='', help_text='Automatically generated on creation', max_length=20, verbose_name='File extension'),
        ),
        migrations.AlterField(
            model_name='element',
            name='filename',
            field=models.CharField(blank=True, db_index=True, default='', help_text='Automatically generated on creation', max_length=255, verbose_name='filename'),
        ),
        migrations.AlterField(
            model_name='element',
            name='filesize',
            field=models.IntegerField(blank=True, db_index=True, default=0, help_text='Automatically generated on creation', verbose_name='File size'),
        ),
        migrations.AlterField(
            model_name='element',
            name='full_name',
            field=models.CharField(blank=True, db_index=True, max_length=255, verbose_name='Complete name'),
        ),
        migrations.AlterField(
            model_name='element',
            name='full_name_normalized',
            field=models.CharField(blank=True, db_index=True, help_text='Complete name without special chars nor accents', max_length=255, verbose_name='Normalized complete name'),
        ),
        migrations.AlterField(
            model_name='element',
            name='long_description',
            field=models.TextField(blank=True, verbose_name='Long description'),
        ),
        migrations.AlterField(
            model_name='element',
            name='md5',
            field=models.CharField(blank=True, db_index=True, default='', help_text='Automatically generated on creation', max_length=120, verbose_name='MD5 sum'),
        ),
        migrations.AlterField(
            model_name='element',
            name='modification',
            field=models.DateTimeField(auto_now=True, db_index=True, verbose_name='Modification date'),
        ),
        migrations.AlterField(
            model_name='element',
            name='name',
            field=models.CharField(db_index=True, max_length=100, verbose_name='Name'),
        ),
        migrations.AlterField(
            model_name='element',
            name='official_link',
            field=models.URLField(blank=True, max_length=255, verbose_name='URL to the web page'),
        ),
        migrations.AlterField(
            model_name='element',
            name='repository',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='repository.Repository', verbose_name='Repository'),
        ),
        migrations.AlterField(
            model_name='element',
            name='sha1',
            field=models.CharField(blank=True, db_index=True, default='', help_text='Automatically generated on creation', max_length=120, verbose_name='SHA1 sum'),
        ),
        migrations.AlterField(
            model_name='element',
            name='sha256',
            field=models.CharField(blank=True, db_index=True, default='', help_text='Automatically generated on creation', max_length=120, verbose_name='SHA256 sum'),
        ),
        migrations.AlterField(
            model_name='element',
            name='short_description',
            field=models.CharField(blank=True, max_length=500, verbose_name='Short description'),
        ),
        migrations.AlterField(
            model_name='element',
            name='slug',
            field=models.SlugField(max_length=100, verbose_name='Slug'),
        ),
        migrations.AlterField(
            model_name='element',
            name='states',
            field=models.ManyToManyField(db_index=True, to='repository.ArchiveState', verbose_name='Archive states'),
        ),
        migrations.AlterField(
            model_name='elementsignature',
            name='creation',
            field=models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Creation date'),
        ),
        migrations.AlterField(
            model_name='elementsignature',
            name='element',
            field=models.ForeignKey(default=0, on_delete=django.db.models.deletion.CASCADE, to='repository.Element', verbose_name='Element'),
        ),
        migrations.AlterField(
            model_name='elementsignature',
            name='method',
            field=models.CharField(choices=[('gpg', 'GnuPG'), ('x509', 'x509 (openSSL/libreSSL)')], db_index=True, max_length=10, verbose_name='signature method'),
        ),
        migrations.AlterField(
            model_name='repository',
            name='admin_group',
            field=models.ManyToManyField(blank=True, db_index=True, related_name='repository_admin', to='auth.Group', verbose_name='Admin groups'),
        ),
        migrations.AlterField(
            model_name='repository',
            name='archive_type',
            field=models.CharField(db_index=True, max_length=100, verbose_name='Repository type'),
        ),
        migrations.AlterField(
            model_name='repository',
            name='author',
            field=models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to=settings.AUTH_USER_MODEL, verbose_name='Author'),
        ),
        migrations.AlterField(
            model_name='repository',
            name='creation',
            field=models.DateTimeField(auto_now_add=True, db_index=True, verbose_name='Creation date'),
        ),
        migrations.AlterField(
            model_name='repository',
            name='is_private',
            field=models.BooleanField(db_index=True, default=False, verbose_name='Should readers be authenticated?'),
        ),
        migrations.AlterField(
            model_name='repository',
            name='modification',
            field=models.DateTimeField(auto_now=True, db_index=True, verbose_name='Modification date'),
        ),
        migrations.AlterField(
            model_name='repository',
            name='name',
            field=models.CharField(db_index=True, max_length=100, verbose_name='Name'),
        ),
        migrations.AlterField(
            model_name='repository',
            name='on_index',
            field=models.BooleanField(db_index=True, default=True, verbose_name='Display on public index?'),
        ),
        migrations.AlterField(
            model_name='repository',
            name='reader_group',
            field=models.ManyToManyField(blank=True, db_index=True, related_name='repository_reader', to='auth.Group', verbose_name='Readers groups'),
        ),
        migrations.AlterField(
            model_name='repository',
            name='slug',
            field=models.SlugField(max_length=100, verbose_name='Slug'),
        ),
    ]



# -*- coding: utf-8 -*-
# Generated by Django 1.11.2 on 2017-11-04 13:44
from __future__ import unicode_literals

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('repository', '0005_auto_20171104_1300'),
    ]

    operations = [
        migrations.AlterModelOptions(
            name='repository',
            options={'verbose_name': 'repository', 'verbose_name_plural': 'repositories'},
        ),
    ]




from django.conf.urls import include, url
from moneta.repository import views


def get_patterns():
    from moneta.repositories.base import RepositoryModelsClasses
    pattern_list_ = []
    for name, model in RepositoryModelsClasses().get_models().items():
        pattern_list_.append(url(r'^p/%s/' % name, (model.public_urls, name, name)))
    pattern_list_.append(url(r'^p/$', views.public_check, name='public_check'))
    for name, model in RepositoryModelsClasses().get_models().items():
        pattern_list_.append(url(r'^a/%s/' % name, (model.urls, 'auth-' + name, 'auth-' + name)))
        pattern_list_.append(url(r'^b/%s/' % name, (model.public_urls, 'authb-' + name, 'authb-' + name)))
    pattern_list_.append(url(r'^a/$', views.private_check, name='private_check_a'))
    pattern_list_.append(url(r'^b/$', views.private_check, name='private_check_b'))
    return pattern_list_


name_set = set()
urlpatterns = get_patterns()
app_name = 'repository'



import os

from djangofloor.log import log_configuration
from moneta.repository.signing import get_gpg

__author__ = 'Matthieu Gallet'


def moneta_log_configuration(settings_dict):
    config = log_configuration(settings_dict)
    config['loggers']['gnupg'] = {'handlers': [], 'level': 'ERROR', 'propagate': True}
    return config


# noinspection PyUnresolvedReferences
moneta_log_configuration.required_settings = log_configuration.required_settings


# noinspection PyPackageRequirements
def auto_generate_signing_key(django_ready):
    if not django_ready:
        return None
    from django.conf import settings
    gnupg_home = settings.GNUPG_HOME
    if not os.path.isdir(gnupg_home):
        os.makedirs(gnupg_home)
    os.chmod(gnupg_home, 0o700)
    for root, dirnames, filenames in os.walk(gnupg_home):
        os.chmod(root, 0o700)
        for filename in filenames:
            os.chmod(os.path.join(root, filename), 0o600)
    gpg = get_gpg()
    if len(gpg.list_keys(False)) == 0:
        input_data = gpg.gen_key_input(key_type='RSA', key_length=2048,
                                       name_real='%s GNUPG key' % settings.SERVER_NAME,
                                       name_comment='GPG key of the %s repository' % settings.SERVER_NAME,
                                       name_email=settings.ADMIN_EMAIL, expire_date='10y')
        print('Generating a new private signing key (can take some time…)')
        gpg.gen_key(input_data)
        print('Private key generated.')
    key_id = None
    for key in gpg.list_keys(False):
        key_id = "{keyid}".format(**key)
    return key_id or ''




from django.contrib import admin
from django.contrib.sites.models import Site

from moneta.repository.models import Element, Repository


class ElementAdmin(admin.ModelAdmin):
    pass

admin.site.unregister(Site)
# admin.site.unregister(SocialToken)
# admin.site.unregister(SocialAccount)
# admin.site.unregister(SocialApp)
# admin.site.unregister(EmailAddress)
# admin.site.unregister(EmailConfirmation)

admin.site.register(Element, ElementAdmin)
admin.site.register(Repository)



"""Setup file for the Moneta project.
"""
import os.path
import re

from setuptools import setup, find_packages

__author__ = 'flanker'


with open(os.path.join(os.path.dirname(__file__), 'README.md'), encoding='utf-8') as fd:
    long_description = fd.read()
version = None
for line in open(os.path.join('moneta', '__init__.py'), 'r', encoding='utf-8'):
    matcher = re.match(r"""^__version__\s*=\s*['"](.*)['"]\s*$""", line)
    version = version or matcher and matcher.group(1)

entry_points = {'console_scripts': ['moneta-ctl = djangofloor.scripts:control']}

setup(
    name='moneta',
    version=version,
    description='Emulator for different kinds (Debian/Yum/Pypi/Maven/…) of package repositories on the same server.',
    long_description=long_description,
    author='flanker',
    author_email='flanker@19pouces.net',
    license='CeCILL-B',
    url='https://moneta.readthedocs.io/',
    entry_points=entry_points,
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    install_requires=['setuptools>=1.0', 'djangofloor>=1.0.25', 'gnupg>=2.3', 'rubymarshal', 'pyyaml'],
    setup_requires=[],
    classifiers=['Development Status :: 5 - Production/Stable',
                 'Framework :: Django :: 1.11',
                 'Framework :: Django :: 2.0',
                 'Natural Language :: English',
                 'Natural Language :: French',
                 'Operating System :: MacOS :: MacOS X',
                 'Operating System :: POSIX :: BSD',
                 'Operating System :: POSIX :: Linux',
                 'Operating System :: Unix',
                 'License :: OSI Approved :: CEA CNRS Inria Logiciel Libre License, version 2.1 (CeCILL-2.1)',
                 'Programming Language :: Python :: 3.5',
                 'Programming Language :: Python :: 3.6',
                 'Programming Language :: Python :: 3.7',
                 'Programming Language :: Python :: 3 :: Only'
                 ],
)



from djangofloor.conf.config_values import Directory, CallableSetting, AutocreateFileContent
from moneta.conf import moneta_log_configuration, auto_generate_signing_key

__author__ = 'flanker'

DF_TEMPLATE_CONTEXT_PROCESSORS = ['moneta.context_processors.context_base']
DF_INDEX_VIEW = 'moneta.views.index'
DF_INSTALLED_APPS = ['moneta', 'moneta.repositories', 'moneta.repository', ]
DF_SITE_SEARCH_VIEW = None
DF_URL_CONF = 'moneta.root_urls.urls'
LISTEN_ADDRESS = '127.0.0.1:8131'
SERVER_BASE_URL = 'http://localhost:8131/'
USE_CELERY = False
UPLOAD_ROOT = 'uploads'
ARCHIVE_FILTERS = [
    'moneta.repository.filters.informations',
    # 'moneta.repository.filters.deb_archive',
]
REPOSITORY_CLASSES = [
    'moneta.repositories.aptitude.Aptitude',
    'moneta.repositories.pypi.Pypi',
    'moneta.repositories.maven3.Maven3',
    'moneta.repositories.flat_files.FlatFile',
    'moneta.repositories.yum.Yum',
    'moneta.repositories.ruby.RubyGem',
    'moneta.repositories.jetbrains.Jetbrains',
    'moneta.repositories.vagrant.Vagrant',
]
USE_HTTP_BASIC_AUTH = True

STORAGES = {
    'archive': {
        'ENGINE': 'moneta.repository.storages.FlatStorage',
        'ROOT': Directory('{MEDIA_ROOT}/archives'),
        'PATH_LEN': 1,
    },
    'default': {
        'ENGINE': 'moneta.repository.storages.FlatStorage',
        'ROOT': Directory('{MEDIA_ROOT}/uncompressed'),
        'PATH_LEN': 1,
    },
    'cache': {
        'ENGINE': 'moneta.repository.storages.FlatStorage',
        'ROOT': Directory('{MEDIA_ROOT}/cache'),
        'PATH_LEN': 1,
    }
}

STORAGE_ARCHIVE = 'archive'
STORAGE_UNCOMPRESSED = 'uncompressed'
STORAGE_CACHE = 'cache'
WEBSOCKET_URL = None

GNUPG_HOME = Directory('{LOCAL_PATH}/gpg')
GNUPG_KEYID = AutocreateFileContent('{LOCAL_PATH}/gpg_key_id.txt', auto_generate_signing_key)
GNUPG_PATH = 'gpg'
GNUPG_DIGEST_ALGO = 'SHA512'

LOGGING = CallableSetting(moneta_log_configuration)
